"""
    Contains classes related to display of data items.
"""

# standard libraries
import copy
import functools
import gettext
import math
import numbers
import operator
import threading
import typing
import uuid
import weakref

import numpy

from nion.data import Calibration
from nion.data import Core
from nion.data import DataAndMetadata
from nion.data import Image
from nion.swift.model import Cache
from nion.swift.model import ColorMaps
from nion.swift.model import Graphics
from nion.utils import Event
from nion.utils import ListModel
from nion.utils import Observable
from nion.utils import Persistence


_ = gettext.gettext


class CalibrationList:

    def __init__(self, calibrations=None):
        self.list = list() if calibrations is None else copy.deepcopy(calibrations)

    def __len__(self):
        return len(self.list)

    def __getitem__(self, item):
        return self.list[item]

    def read_dict(self, storage_list):
        # storage_list will be whatever is returned by write_dict.
        new_list = list()
        for calibration_dict in storage_list:
            new_list.append(Calibration.Calibration().read_dict(calibration_dict))
        self.list = new_list
        return self  # for convenience

    def write_dict(self):
        list = []
        for calibration in self.list:
            list.append(calibration.write_dict())
        return list


class GraphicSelection:
    def __init__(self, indexes=None, anchor_index=None):
        super().__init__()
        self.__changed_event = Event.Event()
        self.__indexes = copy.copy(indexes) if indexes else set()
        self.__anchor_index = anchor_index

    def __copy__(self):
        return type(self)(self.__indexes, self.__anchor_index)

    def __eq__(self, other):
        return other is not None and self.indexes == other.indexes and self.anchor_index == other.anchor_index

    def __ne__(self, other):
        return other is None or self.indexes != other.indexes or self.anchor_index != other.anchor_index

    @property
    def changed_event(self):
        return self.__changed_event

    @property
    def current_index(self):
        if len(self.__indexes) == 1:
            for index in self.__indexes:
                return index
        return None

    @property
    def anchor_index(self):
        return self.__anchor_index

    @property
    def has_selection(self):
        return len(self.__indexes) > 0

    def contains(self, index):
        return index in self.__indexes

    @property
    def indexes(self):
        return self.__indexes

    def clear(self):
        old_index = self.__indexes.copy()
        self.__indexes = set()
        self.__anchor_index = None
        if old_index != self.__indexes:
            self.__changed_event.fire()

    def __update_anchor_index(self):
        for index in self.__indexes:
            if self.__anchor_index is None or index < self.__anchor_index:
                self.__anchor_index = index

    def add(self, index):
        assert isinstance(index, numbers.Integral)
        old_index = self.__indexes.copy()
        self.__indexes.add(index)
        if len(old_index) == 0:
            self.__anchor_index = index
        if old_index != self.__indexes:
            self.__changed_event.fire()

    def remove(self, index):
        assert isinstance(index, numbers.Integral)
        old_index = self.__indexes.copy()
        self.__indexes.remove(index)
        if not self.__anchor_index in self.__indexes:
            self.__update_anchor_index()
        if old_index != self.__indexes:
            self.__changed_event.fire()

    def add_range(self, range):
        for index in range:
            self.add(index)

    def set(self, index):
        assert isinstance(index, numbers.Integral)
        old_index = self.__indexes.copy()
        self.__indexes = set()
        self.__indexes.add(index)
        self.__anchor_index = index
        if old_index != self.__indexes:
            self.__changed_event.fire()

    def toggle(self, index):
        assert isinstance(index, numbers.Integral)
        if index in self.__indexes:
            self.remove(index)
        else:
            self.add(index)

    def insert_index(self, new_index):
        new_indexes = set()
        for index in self.__indexes:
            if index < new_index:
                new_indexes.add(index)
            else:
                new_indexes.add(index + 1)
        if self.__anchor_index is not None:
            if new_index <= self.__anchor_index:
                self.__anchor_index += 1
        if self.__indexes != new_indexes:
            self.__indexes = new_indexes
            self.changed_event.fire()

    def remove_index(self, remove_index):
        new_indexes = set()
        for index in self.__indexes:
            if index != remove_index:
                if index > remove_index:
                    new_indexes.add(index - 1)
                else:
                    new_indexes.add(index)
        if self.__anchor_index is not None:
            if remove_index == self.__anchor_index:
                self.__update_anchor_index()
            elif remove_index < self.__anchor_index:
                self.__anchor_index -= 1
        if self.__indexes != new_indexes:
            self.__indexes = new_indexes
            self.changed_event.fire()


def calculate_display_range(display_limits, data_range, data_sample, xdata, complex_display_type):
    if display_limits is not None:
        display_limit_low = display_limits[0] if display_limits[0] is not None else data_range[0]
        display_limit_high = display_limits[1] if display_limits[1] is not None else data_range[1]
        return display_limit_low, display_limit_high
    if xdata and xdata.is_data_complex_type and complex_display_type is None:  # log absolute
        if data_sample is not None:
            fraction = 0.05
            display_limit_low = data_sample[int(data_sample.shape[0] * fraction)]
            display_limit_high = data_range[1]
            return display_limit_low, display_limit_high
    return data_range


class CalibrationStyle:
    label = None
    calibration_style_id = None
    is_calibrated = False

    def get_dimensional_calibrations(self, dimensional_shape, dimensional_calibrations) -> typing.Sequence[Calibration.Calibration]:
        return list()

    def get_intensity_calibration(self, xdata: DataAndMetadata.DataAndMetadata) -> Calibration.Calibration:
        return xdata.intensity_calibration if self.is_calibrated else Calibration.Calibration()


class CalibrationStyleNative(CalibrationStyle):
    label = _("Calibrated")
    calibration_style_id = "calibrated"
    is_calibrated = True

    def get_dimensional_calibrations(self, dimensional_shape, dimensional_calibrations) -> typing.Sequence[Calibration.Calibration]:
        return dimensional_calibrations


class CalibrationStylePixelsTopLeft(CalibrationStyle):
    label = _("Pixels (Top-Left)")
    calibration_style_id = "pixels-top-left"

    def get_dimensional_calibrations(self, dimensional_shape, dimensional_calibrations) -> typing.Sequence[Calibration.Calibration]:
        return [Calibration.Calibration() for display_dimension in dimensional_shape]


class CalibrationStylePixelsCenter(CalibrationStyle):
    label = _("Pixels (Center)")
    calibration_style_id = "pixels-center"

    def get_dimensional_calibrations(self, dimensional_shape, dimensional_calibrations) -> typing.Sequence[Calibration.Calibration]:
        return [Calibration.Calibration(offset=-display_dimension/2) for display_dimension in dimensional_shape]


class CalibrationStyleFractionalTopLeft(CalibrationStyle):
    label = _("Fractional (Top Left)")
    calibration_style_id = "relative-top-left"

    def get_dimensional_calibrations(self, dimensional_shape, dimensional_calibrations) -> typing.Sequence[Calibration.Calibration]:
        return [Calibration.Calibration(scale=1.0/display_dimension) for display_dimension in dimensional_shape]


class CalibrationStyleFractionalCenter(CalibrationStyle):
    label = _("Fractional (Center)")
    calibration_style_id = "relative-center"

    def get_dimensional_calibrations(self, dimensional_shape, dimensional_calibrations) -> typing.Sequence[Calibration.Calibration]:
        return [Calibration.Calibration(scale=2.0/display_dimension, offset=-1.0) for display_dimension in dimensional_shape]


class DisplayValues:
    """Display data used to render the display."""

    def __init__(self, data_and_metadata, sequence_index, collection_index, slice_center, slice_width, display_limits, complex_display_type, color_map_data):
        self.__lock = threading.RLock()
        self.__data_and_metadata = data_and_metadata
        self.__sequence_index = sequence_index
        self.__collection_index = collection_index
        self.__slice_center = slice_center
        self.__slice_width = slice_width
        self.__display_limits = display_limits
        self.__complex_display_type = complex_display_type
        self.__color_map_data = color_map_data
        self.__display_data_and_metadata_dirty = True
        self.__display_data_and_metadata = None
        self.__data_range_dirty = True
        self.__data_range = None
        self.__data_sample_dirty = True
        self.__data_sample = None
        self.__display_range_dirty = True
        self.__display_range = None
        self.__display_rgba_dirty = True
        self.__display_rgba = None
        self.__display_rgba_timestamp = data_and_metadata.timestamp if data_and_metadata else None
        self.__finalized = False
        self.on_finalize = None

    def finalize(self):
        with self.__lock:
            self.__finalized = True
        if callable(self.on_finalize):
            self.on_finalize(self)

    @property
    def color_map_data(self):
        return self.__color_map_data

    @property
    def display_data_and_metadata(self):
        with self.__lock:
            if self.__display_data_and_metadata_dirty:
                self.__display_data_and_metadata_dirty = False
                data_and_metadata = self.__data_and_metadata
                if data_and_metadata is not None:
                    timestamp = data_and_metadata.timestamp
                    data_and_metadata, modified = Core.function_display_data_no_copy(data_and_metadata, self.__sequence_index, self.__collection_index, self.__slice_center, self.__slice_width, self.__complex_display_type)
                    if data_and_metadata:
                        data_and_metadata.data_metadata.timestamp = timestamp
                    self.__display_data_and_metadata = data_and_metadata
            return self.__display_data_and_metadata

    @property
    def data_range(self):
        with self.__lock:
            if self.__data_range_dirty:
                self.__data_range_dirty = False
                display_data_and_metadata = self.display_data_and_metadata
                display_data = display_data_and_metadata.data if display_data_and_metadata else None
                if display_data is not None and display_data.size and self.__data_and_metadata:
                    data_shape = self.__data_and_metadata.data_shape
                    data_dtype = self.__data_and_metadata.data_dtype
                    if Image.is_shape_and_dtype_rgb_type(data_shape, data_dtype):
                        self.__data_range = (0, 255)
                    elif Image.is_shape_and_dtype_complex_type(data_shape, data_dtype):
                        self.__data_range = (numpy.amin(display_data), numpy.amax(display_data))
                    else:
                        self.__data_range = (numpy.amin(display_data), numpy.amax(display_data))
                else:
                    self.__data_range = None
                if self.__data_range is not None:
                    if math.isnan(self.__data_range[0]) or math.isnan(self.__data_range[1]) or math.isinf(self.__data_range[0]) or math.isinf(self.__data_range[1]):
                        self.__data_range = (0.0, 0.0)
            return self.__data_range

    @property
    def display_range(self):
        with self.__lock:
            if self.__display_range_dirty:
                self.__display_range_dirty = False
                self.__display_range = calculate_display_range(self.__display_limits, self.data_range, self.data_sample, self.__data_and_metadata, self.__complex_display_type)
            return self.__display_range

    @property
    def data_sample(self):
        with self.__lock:
            if self.__data_sample_dirty:
                self.__data_sample_dirty = False
                display_data_and_metadata = self.display_data_and_metadata
                display_data = display_data_and_metadata.data if display_data_and_metadata else None
                if display_data is not None and display_data.size and self.__data_and_metadata:
                    data_shape = self.__data_and_metadata.data_shape
                    data_dtype = self.__data_and_metadata.data_dtype
                    if Image.is_shape_and_dtype_rgb_type(data_shape, data_dtype):
                        self.__data_sample = None
                    elif Image.is_shape_and_dtype_complex_type(data_shape, data_dtype):
                        self.__data_sample = numpy.sort(numpy.random.choice(display_data.reshape(numpy.product(display_data.shape)), 200))
                    else:
                        self.__data_sample = None
                else:
                    self.__data_sample = None
            return self.__data_sample

    @property
    def display_rgba(self):
        with self.__lock:
            if self.__display_rgba_dirty:
                self.__display_rgba_dirty = False
                display_data_and_metadata = self.display_data_and_metadata
                if display_data_and_metadata is not None and self.__data_and_metadata is not None:
                    if self.data_range is not None:  # workaround until validating and retrieving data stats is an atomic operation
                        # display_range is just display_limits but calculated if display_limits is None
                        display_range = self.display_range
                        self.__display_rgba = Core.function_display_rgba(display_data_and_metadata, display_range, self.__color_map_data).data
            return self.__display_rgba

    @property
    def display_rgba_timestamp(self):
        return self.__display_rgba_timestamp


class ObservableListEventObserver:

    def __init__(self, container, key, event, fn):
        self.__event_listeners = list()

        def item_inserted(key_, value, index):
            if key == key_:
                def handle_event(*args, **kwargs):
                    fn(value, *args, **kwargs)

                self.__event_listeners.insert(index, getattr(value, event).listen(handle_event))

        def item_removed(key_, value, index):
            if key_ == key:
                self.__event_listeners[index].close()
                del self.__event_listeners[index]

        self.__item_inserted_event_listener = container.item_inserted_event.listen(item_inserted)
        self.__item_removed_event_listener = container.item_removed_event.listen(item_removed)

        for index, item in enumerate(getattr(container, key)):
            item_inserted(key, item, index)

    def close(self):
        self.__item_inserted_event_listener.close()
        self.__item_inserted_event_listener = None
        self.__item_removed_event_listener.close()
        self.__item_removed_event_listener = None
        for event_listener in self.__event_listeners:
            event_listener.close()
        self.__event_listeners = None


class Display(Observable.Observable, Persistence.PersistentObject):
    """The display properties for a DataItem.

    Also handles conversion of raw data to formats suitable for display such as raster RGBA.

    Display data is the associated data item data after it has been reduced to basic form (1d, 2d) by slicing and/or
    conversion to scalar.

    RGB data is the display data after it has been converted to RGB for raster display.

    In order to be able to efficiently route data in a thread, and also to make the data consistent with a snapshot
    in time, display data and RGB data is managed via a DisplayValues object.

    In addition to regular observable events, this class also generates the following events:
        - about_to_be_removed_event: fired when about to be removed from parent container.
        - display_changed_event: fired when display changes in a way to affect drawing.
        - display_data_will_change_event: fired when display data changes.
        - display_type_changed_event: fired when the display type changes.
        - display_graphic_selection_changed_event: fired when the graphic selection changes.
    """

    def __init__(self):
        super().__init__()
        self.__container_weak_ref = None
        self.__cache = Cache.ShadowCache()
        self.__color_map_data = None
        # display type to use
        self.define_property("display_type", changed=self.__display_type_changed)
        # conversion to scalar
        self.define_property("complex_display_type", changed=self.__property_changed)
        # calibration display
        self.define_property("display_calibrated_values", True, changed=self.__property_changed)
        self.define_property("dimensional_calibration_style", None, changed=self.__property_changed)
        # data scaling and color (raster)
        self.define_property("display_limits", validate=self.__validate_display_limits, changed=self.__property_changed)
        self.define_property("color_map_id", changed=self.__color_map_id_changed)
        # image zoom and position
        self.define_property("image_zoom", 1.0, changed=self.__property_changed)
        self.define_property("image_position", (0.5, 0.5), changed=self.__property_changed)
        self.define_property("image_canvas_mode", "fit", changed=self.__property_changed)
        # line plot axes and labels
        self.define_property("y_min", changed=self.__property_changed)
        self.define_property("y_max", changed=self.__property_changed)
        self.define_property("y_style", "linear", changed=self.__property_changed)
        self.define_property("left_channel", changed=self.__property_changed)
        self.define_property("right_channel", changed=self.__property_changed)
        self.define_property("legend_labels", changed=self.__property_changed)
        self.define_property("intensity_calibration", None, make=Calibration.Calibration, changed=self.__property_changed)
        self.define_property("dimensional_calibrations", CalibrationList(), hidden=True, make=CalibrationList, changed=self.__property_changed)
        self.define_property("dimensional_scales", None, changed=self.__property_changed)
        # slicing data to 1d or 2d
        self.define_property("sequence_index", 0, validate=self.__validate_sequence_index, changed=self.__property_changed)
        self.define_property("collection_index", (0, 0, 0), validate=self.__validate_collection_index, changed=self.__property_changed)
        self.define_property("slice_center", 0, validate=self.__validate_slice_center, changed=self.__slice_interval_changed)
        self.define_property("slice_width", 1, validate=self.__validate_slice_width, changed=self.__slice_interval_changed)
        # display script
        self.define_property("display_script", changed=self.__property_changed)
        # graphics
        self.define_relationship("graphics", Graphics.factory, insert=self.__insert_graphic, remove=self.__remove_graphic)

        self.__title = None  # not persistent during display panel transition

        self.item_changed_event = Event.Event()  # for indicated this display has mutated somehow

        # # last display values is the last one to be fully displayed.
        # # when the current display values makes it all the way to display, it will fire an event.
        # # the display will listen for that event and update last display values.
        self.__last_display_values = None
        self.__current_display_values = None
        self.__is_master = True

        self.display_values_changed_event = Event.Event()
        self.__calculated_display_values_available_event = Event.Event()

        self.__graphics_map = dict()  # type: typing.MutableMapping[uuid.UUID, Graphics.Graphic]
        self.__graphic_changed_listeners = list()
        self.__data_and_metadata = None  # the most recent data to be displayed. should have immediate data available.
        self.graphic_selection = GraphicSelection()

        def graphic_selection_changed():
            # relay the message
            self.display_graphic_selection_changed_event.fire(self.graphic_selection)

        self.__graphic_selection_changed_event_listener = self.graphic_selection.changed_event.listen(graphic_selection_changed)
        self.about_to_be_removed_event = Event.Event()
        self.display_changed_event = Event.Event()
        self.display_data_will_change_event = Event.Event()
        self.display_type_changed_event = Event.Event()
        self.display_graphic_selection_changed_event = Event.Event()
        self._about_to_be_removed = False
        self._closed = False

    def close(self):
        self.__graphic_selection_changed_event_listener.close()
        self.__graphic_selection_changed_event_listener = None
        for graphic in copy.copy(self.graphics):
            self.__disconnect_graphic(graphic, 0)
            graphic.close()
        self.graphic_selection = None
        assert self._about_to_be_removed
        assert not self._closed
        self._closed = True
        self.__container_weak_ref = None

    def read_from_dict(self, properties):
        super().read_from_dict(properties)
        dimensional_calibration_style_property = self._get_persistent_property("dimensional_calibration_style")
        if self.dimensional_calibration_style is None:
            calibration_style = self.default_calibrated_calibration_style if self.display_calibrated_values else self.default_uncalibrated_calibration_style
            dimensional_calibration_style_property.value = calibration_style.calibration_style_id
        else:
            valid_calibration_styles = (calibrated_style.calibration_style_id for calibrated_style in self.calibration_styles)
            if not dimensional_calibration_style_property.value in valid_calibration_styles:
                calibration_style = self.default_calibrated_calibration_style if dimensional_calibration_style_property.value.startswith("calibration") else self.default_uncalibrated_calibration_style
                dimensional_calibration_style_property.value = calibration_style.calibration_style_id

    def save_properties(self) -> typing.Tuple:
        return (
            self.left_channel,
            self.right_channel,
            self.y_min,
            self.y_max,
            self.y_style,
            self.image_zoom,
            self.image_position,
            self.image_canvas_mode,
            self.complex_display_type,
            self.display_calibrated_values,
            self.dimensional_calibration_style,
            self.display_limits,
            self.color_map_id,
            self.sequence_index,
            self.collection_index,
            self.slice_center,
            self.slice_interval,
            self.display_type,
            self.display_script
        )

    def restore_properties(self, properties: typing.Tuple) -> None:
        self.left_channel = properties[0]
        self.right_channel = properties[1]
        self.y_min = properties[2]
        self.y_max = properties[3]
        self.y_style = properties[4]
        self.image_zoom = properties[5]
        self.image_position = properties[6]
        self.image_canvas_mode = properties[7]
        self.complex_display_type = properties[8]
        self.display_calibrated_values = properties[9]
        self.dimensional_calibration_style = properties[10]
        self.display_limits = properties[11]
        self.color_map_id = properties[12]
        self.sequence_index = properties[13]
        self.collection_index = properties[14]
        self.slice_center = properties[15]
        self.slice_interval = properties[16]
        self.display_type = properties[17]
        self.display_script = properties[18]

    @property
    def container(self):
        return self.__container_weak_ref()

    def about_to_be_inserted(self, container):
        assert self.__container_weak_ref is None
        self.__container_weak_ref = weakref.ref(container)

        # child displays model will watch the container for data items and make a new list with each of the displays in each data item.
        self.__child_displays_model = ListModel.FlattenedListModel(container=container, master_items_key="data_items", child_items_key="displays")

        def handle_display_changed(display):
            self.display_changed_event.fire()

        # watch the child display model for insert/removes from displays; and passes on the display_changed_event for each active display
        self.__child_display_observer = ObservableListEventObserver(self.__child_displays_model, "displays", "display_changed_event", handle_display_changed)

    def about_to_be_removed(self):
        # called before close and before item is removed from its container
        for graphic in self.graphics:
            graphic.about_to_be_removed()
        self.about_to_be_removed_event.fire()
        self.__child_display_observer.close()
        self.__child_display_observer = None
        self.__child_displays_model.close()
        self.__child_displays_model = None
        assert not self._about_to_be_removed
        self._about_to_be_removed = True

    def insert_model_item(self, container, name, before_index, item):
        if self.__container_weak_ref:
            self.container.insert_model_item(container, name, before_index, item)
        else:
            container.insert_item(name, before_index, item)

    def remove_model_item(self, container, name, item, *, safe: bool=False) -> typing.Optional[typing.Sequence]:
        if self.__container_weak_ref:
            return self.container.remove_model_item(container, name, item, safe=safe)
        else:
            container.remove_item(name, item)
            return None

    def clone(self) -> "Display":
        display = Display()
        display.uuid = self.uuid
        for graphic in self.graphics:
            display.add_graphic(graphic.clone())
        return display

    @property
    def _display_cache(self):
        return self.__cache

    @property
    def child_displays_model(self):
        return self.__child_displays_model

    @property
    def title(self) -> str:
        return self.__title

    @title.setter
    def title(self, value: str) -> None:
        self.__title = value
        self.notify_property_changed("title")

    @property
    def dimensional_calibrations(self) -> typing.Sequence[Calibration.Calibration]:
        return copy.deepcopy(self._get_persistent_property_value("dimensional_calibrations", CalibrationList()).list)

    @dimensional_calibrations.setter
    def dimensional_calibrations(self, dimensional_calibrations: typing.Sequence[Calibration.Calibration]) -> None:
        """ Set the dimensional calibrations. """
        self._set_persistent_property_value("dimensional_calibrations", CalibrationList(dimensional_calibrations))

    # when the data item changes, it will call this method so that item_changed can be fired.
    def _item_changed(self):
        self.item_changed_event.fire()

    @property
    def data_and_metadata_for_display_panel(self):
        return self.__data_and_metadata

    def reset_display_limits(self):
        """Reset display limits so that they are auto calculated whenever the data changes."""
        self.display_limits = None

    def auto_display_limits(self):
        """Calculate best display limits and set them."""
        display_data_and_metadata = self.get_calculated_display_values(True).display_data_and_metadata
        data = display_data_and_metadata.data if display_data_and_metadata else None
        if data is not None:
            # The old algorithm was a problem during EELS where the signal data
            # is a small percentage of the overall data and was falling outside
            # the included range. This is the new simplified algorithm. Future
            # feature may allow user to select more complex algorithms.
            mn, mx = numpy.nanmin(data), numpy.nanmax(data)
            self.display_limits = mn, mx

    def view_to_intervals(self, data_and_metadata: DataAndMetadata.DataAndMetadata, intervals: typing.List[typing.Tuple[float, float]]) -> None:
        """Change the view to encompass the channels and data represented by the given intervals."""
        left = None
        right = None
        for interval in intervals:
            left = min(left, interval[0]) if left is not None else interval[0]
            right = max(right, interval[1]) if right is not None else interval[1]
        left = left if left is not None else 0.0
        right = right if right is not None else 1.0
        extra = (right - left) * 0.5
        self.left_channel = int(max(0.0, left - extra) * data_and_metadata.data_shape[-1])
        self.right_channel = int(min(1.0, right + extra) * data_and_metadata.data_shape[-1])
        data_min = numpy.amin(data_and_metadata.data[..., self.left_channel:self.right_channel])
        data_max = numpy.amax(data_and_metadata.data[..., self.left_channel:self.right_channel])
        if data_min > 0 and data_max > 0:
            self.y_min = 0.0
            self.y_max = data_max * 1.2
        elif data_min < 0 and data_max < 0:
            self.y_min = data_min * 1.2
            self.y_max = 0.0
        else:
            self.y_min = data_min * 1.2
            self.y_max = data_max * 1.2

    def view_to_selected_graphics(self, data_and_metadata: DataAndMetadata.DataAndMetadata) -> None:
        """Change the view to encompass the selected graphic intervals."""
        all_graphics = self.graphics
        graphics = [graphic for graphic_index, graphic in enumerate(all_graphics) if self.graphic_selection.contains(graphic_index)]
        intervals = list()
        for graphic in graphics:
            if isinstance(graphic, Graphics.IntervalGraphic):
                intervals.append(graphic.interval)
        self.view_to_intervals(data_and_metadata, intervals)

    @property
    def preview_2d_shape(self) -> typing.Optional[typing.Tuple[int, ...]]:
        if not self.__data_and_metadata:
            return None
        data_and_metadata = self.__data_and_metadata
        dimensional_shape = data_and_metadata.dimensional_shape
        next_dimension = 0
        if data_and_metadata.is_sequence:
            next_dimension += 1
        if data_and_metadata.is_collection:
            collection_dimension_count = data_and_metadata.collection_dimension_count
            datum_dimension_count = data_and_metadata.datum_dimension_count
            # next dimensions are treated as collection indexes.
            if collection_dimension_count == 1 and datum_dimension_count == 1:
                return dimensional_shape[next_dimension:next_dimension + collection_dimension_count + datum_dimension_count]
            elif collection_dimension_count == 2 and datum_dimension_count == 1:
                return dimensional_shape[next_dimension:next_dimension + collection_dimension_count]
            else:  # default, "pick"
                return dimensional_shape[next_dimension + collection_dimension_count:next_dimension + collection_dimension_count + datum_dimension_count]
        else:
            return dimensional_shape[next_dimension:]

    @property
    def selected_graphics(self):
        return [self.graphics[i] for i in self.graphic_selection.indexes]

    def __validate_display_limits(self, value):
        if value is not None:
            if len(value) == 0:
                return None
            elif len(value) == 1:
                return (value[0], None) if value[0] is not None else None
            elif value[0] is not None and value[1] is not None:
                return min(value[0], value[1]), max(value[0], value[1])
            elif value[0] is None and value[1] is None:
                return None
            else:
                return value[0], value[1]
        return value

    def __validate_sequence_index(self, value: int) -> int:
        if not self._is_reading:
            if self.__data_and_metadata and self.__data_and_metadata.dimensional_shape is not None:
                return max(min(int(value), self.__data_and_metadata.max_sequence_index - 1), 0) if self.__data_and_metadata.is_sequence else 0
        return 0

    def __validate_collection_index(self, value: typing.Tuple[int, int, int]) -> typing.Tuple[int, int, int]:
        if not self._is_reading:
            if self.__data_and_metadata and self.__data_and_metadata.dimensional_shape is not None:
                dimensional_shape = self.__data_and_metadata.dimensional_shape
                collection_base_index = 1 if self.__data_and_metadata.is_sequence else 0
                collection_dimension_count = self.__data_and_metadata.collection_dimension_count
                i0 = max(min(int(value[0]), dimensional_shape[collection_base_index + 0] - 1), 0) if collection_dimension_count > 0 else 0
                i1 = max(min(int(value[1]), dimensional_shape[collection_base_index + 1] - 1), 0) if collection_dimension_count > 1 else 0
                i2 = max(min(int(value[2]), dimensional_shape[collection_base_index + 2] - 1), 0) if collection_dimension_count > 2 else 0
                return i0, i1, i2
        return (0, 0, 0)

    def __validate_slice_center_for_width(self, value, slice_width):
        if self.__data_and_metadata and self.__data_and_metadata.dimensional_shape is not None:
            depth = self.__data_and_metadata.dimensional_shape[-1]
            mn = max(int(slice_width * 0.5), 0)
            mx = min(int(depth - slice_width * 0.5), depth - 1)
            return min(max(int(value), mn), mx)
        return value if self._is_reading else 0

    def __validate_slice_center(self, value):
        return self.__validate_slice_center_for_width(value, self.slice_width)

    def __validate_slice_width(self, value):
        if self.__data_and_metadata and self.__data_and_metadata.dimensional_shape is not None:
            depth = self.__data_and_metadata.dimensional_shape[-1]  # signal_index
            slice_center = self.slice_center
            mn = 1
            mx = max(min(slice_center, depth - slice_center) * 2, 1)
            return min(max(value, mn), mx)
        return value if self._is_reading else 1

    def validate_slice_indexes(self) -> None:
        sequence_index = self.__validate_sequence_index(self.sequence_index)
        if sequence_index != self.sequence_index:
            self.sequence_index = sequence_index

        collection_index = self.__validate_collection_index(self.collection_index)
        if collection_index != self.collection_index:
            self.collection_index = collection_index

        slice_center = self.__validate_slice_center_for_width(self.slice_center, 1)
        if slice_center != self.slice_center:
            old_slice_width = self.slice_width
            self.slice_width = 1
            self.slice_center = self.slice_center
            self.slice_width = old_slice_width

    @property
    def actual_display_type(self):
        display_type = self.display_type
        data_and_metadata = self.__data_and_metadata
        valid_data = functools.reduce(operator.mul, self.preview_2d_shape) > 0 if self.preview_2d_shape is not None else False
        if valid_data and data_and_metadata and not display_type in ("line_plot", "image", "display_script"):
            if data_and_metadata.collection_dimension_count == 2 and data_and_metadata.datum_dimension_count == 1:
                display_type = "image"
            elif data_and_metadata.datum_dimension_count == 1:
                display_type = "line_plot"
            elif data_and_metadata.datum_dimension_count == 2:
                display_type = "image"
            # override
            if self.display_script:
                display_type = "display_script"
        if not valid_data and len(self.__child_displays_model.items) > 0:
            if display_type not in ("line_plot", ):
                display_type = "composite-image"
        return display_type

    @property
    def slice_interval(self):
        if self.__data_and_metadata and self.__data_and_metadata.dimensional_shape is not None:
            depth = self.__data_and_metadata.dimensional_shape[-1]  # signal_index
            if depth > 0:
                slice_interval_start = round(self.slice_center - self.slice_width * 0.5)
                slice_interval_end = slice_interval_start + self.slice_width
                return (float(slice_interval_start) / depth, float(slice_interval_end) / depth)
            return 0, 0
        return None

    @slice_interval.setter
    def slice_interval(self, slice_interval):
        if self.__data_and_metadata.dimensional_shape is not None:
            depth = self.__data_and_metadata.dimensional_shape[-1]  # signal_index
            if depth > 0:
                slice_interval_center = round(((slice_interval[0] + slice_interval[1]) * 0.5) * depth)
                slice_interval_width = round((slice_interval[1] - slice_interval[0]) * depth)
                self.slice_center = slice_interval_center
                self.slice_width = slice_interval_width

    def __slice_interval_changed(self, name, value):
        # notify for dependent slice_interval property
        self.__property_changed(name, value)
        self.notify_property_changed("slice_interval")

    def __display_type_changed(self, property_name, value):
        self.__property_changed(property_name, value)
        self.display_type_changed_event.fire()

    def __color_map_id_changed(self, property_name, value):
        self.__property_changed(property_name, value)
        if value:
            lookup_table_options = ColorMaps.color_maps
            self.__color_map_data = lookup_table_options.get(value)
        else:
            self.__color_map_data = None
        self.__property_changed("color_map_data", self.__color_map_data)

    @property
    def color_map_data(self) -> typing.Optional[numpy.ndarray]:
        """Return the color map data as a uint8 ndarray with shape (256, 3)."""
        if self.preview_2d_shape is None:  # is there display data?
            return None
        else:
            return self.__color_map_data if self.__color_map_data is not None else ColorMaps.color_maps.get("grayscale")

    def __property_changed(self, property_name, value):
        # when one of the defined properties changes, this gets called
        self.notify_property_changed(property_name)
        self.display_changed_event.fire()
        if property_name in ("sequence_index", "collection_index", "slice_center", "slice_width", "complex_display_type", "display_limits", "color_map_data"):
            self.display_data_will_change_event.fire()
            self.__send_next_calculated_display_values()
        if property_name in ("dimensional_calibration_style", ):
            self.notify_property_changed("displayed_dimensional_scales")
            self.notify_property_changed("displayed_dimensional_calibrations")
            self.notify_property_changed("displayed_intensity_calibration")
            calibration_style = self.__get_calibration_style_for_id(value)
            self._get_persistent_property("display_calibrated_values").value = calibration_style.is_calibrated if calibration_style else None
        if property_name in ("dimensional_calibrations", "intensity_calibration", "dimensional_scales"):
            self.notify_property_changed("displayed_dimensional_scales")
            self.notify_property_changed("displayed_dimensional_calibrations")
            self.notify_property_changed("displayed_intensity_calibration")

    # message sent when data changes.
    # thread safe
    def update_data(self, data_and_metadata):
        old_data_shape = self.__data_and_metadata.data_shape if self.__data_and_metadata else None
        self.__data_and_metadata = data_and_metadata
        new_data_shape = self.__data_and_metadata.data_shape if self.__data_and_metadata else None
        if old_data_shape != new_data_shape:
            self.validate_slice_indexes()
        self.__send_next_calculated_display_values()
        self.notify_property_changed("displayed_dimensional_calibrations")
        self.notify_property_changed("displayed_intensity_calibration")
        self.display_changed_event.fire()

    def set_storage_cache(self, storage_cache):
        self.__cache.set_storage_cache(storage_cache, self)

    def add_calculated_display_values_listener(self, callback, send=True):
        listener = self.__calculated_display_values_available_event.listen(callback)
        if send:
            self.__send_next_calculated_display_values()
        return listener

    def __send_next_calculated_display_values(self) -> None:
        """Fire event to signal new display values are available."""
        self.__current_display_values = None
        self.__calculated_display_values_available_event.fire()

    def get_calculated_display_values(self, immediate: bool=False) -> DisplayValues:
        """Return the display values.

        Return the current (possibly uncalculated) display values unless 'immediate' is specified.

        If 'immediate', return the existing (calculated) values if they exist. Using the 'immediate' values
        avoids calculation except in cases where the display values haven't already been calculated.
        """
        if not immediate or not self.__is_master or not self.__last_display_values:
            if not self.__current_display_values:
                self.__current_display_values = DisplayValues(self.__data_and_metadata, self.sequence_index, self.collection_index, self.slice_center, self.slice_width, self.display_limits, self.complex_display_type, self.__color_map_data)

                def finalize(display_values):
                    self.__last_display_values = display_values
                    self.display_values_changed_event.fire()

                self.__current_display_values.on_finalize = finalize
            return self.__current_display_values
        return self.__last_display_values

    def increment_display_ref_count(self, amount: int=1):
        self.container.increment_display_ref_count(amount)

    def decrement_display_ref_count(self, amount: int=1):
        self.container.decrement_display_ref_count(amount)

    def _become_master(self):
        self.__is_master = True

    def _relinquish_master(self):
        self.__is_master = False

    def _changes(self):
        return self.container.data_item_changes()

    @property
    def is_live(self):
        return self.container.is_live

    def __insert_graphic(self, name, before_index, graphic):
        graphic.about_to_be_inserted(self)
        graphic_changed_listener = graphic.graphic_changed_event.listen(functools.partial(self.graphic_changed, graphic))
        self.__graphic_changed_listeners.insert(before_index, graphic_changed_listener)
        self.__graphics_map[graphic.uuid] = graphic
        self.graphic_selection.insert_index(before_index)
        self.display_changed_event.fire()
        self.notify_insert_item("graphics", graphic, before_index)

    def __remove_graphic(self, name, index, graphic):
        graphic.about_to_be_removed()
        self.__graphics_map.pop(graphic.uuid)
        self.__disconnect_graphic(graphic, index)
        graphic.close()

    def __disconnect_graphic(self, graphic, index):
        graphic_changed_listener = self.__graphic_changed_listeners[index]
        graphic_changed_listener.close()
        self.__graphic_changed_listeners.remove(graphic_changed_listener)
        self.graphic_selection.remove_index(index)
        self.display_changed_event.fire()
        self.notify_remove_item("graphics", graphic, index)

    def insert_graphic(self, before_index, graphic):
        """Insert a graphic before the index, but do it through the container, so dependencies can be tracked."""
        self.insert_model_item(self, "graphics", before_index, graphic)

    def add_graphic(self, graphic):
        """Append a graphic, but do it through the container, so dependencies can be tracked."""
        self.insert_model_item(self, "graphics", self.item_count("graphics"), graphic)

    def remove_graphic(self, graphic: Graphics.Graphic, *, safe: bool=False) -> typing.Optional[typing.Sequence]:
        """Remove a graphic, but do it through the container, so dependencies can be tracked."""
        return self.remove_model_item(self, "graphics", graphic, safe=safe)

    def get_graphic_by_uuid(self, graphic_uuid: uuid.UUID) -> Graphics.Graphic:
        return self.__graphics_map.get(graphic_uuid)

    # this message comes from the graphic. the connection is established when a graphic
    # is added or removed from this object.
    def graphic_changed(self, graphic):
        self.display_changed_event.fire()

    @property
    def displayed_dimensional_scales(self) -> typing.Sequence[float]:
        """The scale of the fractional coordinate system.

        For displays associated with a single data item, this matches the size of the data.

        For displays associated with a composite data item, this must be stored in this class.
        """
        if self.__data_and_metadata:
            return self.__data_and_metadata.dimensional_shape
        if self.dimensional_scales:
            return self.dimensional_scales
        return [1, 1]

    def get_dimensional_calibrations_with_calibration_style(self, dimensional_calibration_style) -> typing.Sequence[Calibration.Calibration]:
        calibration_style = self.__get_calibration_style_for_id(dimensional_calibration_style)
        if self.__data_and_metadata:
            calibration_style = CalibrationStyleNative() if calibration_style is None else calibration_style
            return calibration_style.get_dimensional_calibrations(self.__data_and_metadata.dimensional_shape, self.__data_and_metadata.dimensional_calibrations)
        else:
            calibration_style = CalibrationStyleNative() if calibration_style is None else calibration_style
            dimensional_scales = self.dimensional_scales
            dimensional_scales = [1] if dimensional_scales is None and self.display_type == "line_plot" else dimensional_scales
            dimensional_scales = [1, 1] if dimensional_scales is None and self.display_type == "image" else dimensional_scales
            dimensional_calibrations = self.dimensional_calibrations if self.dimensional_calibrations else [Calibration.Calibration() for i in range(len(dimensional_scales))]
            return calibration_style.get_dimensional_calibrations(dimensional_scales, dimensional_calibrations)

    @property
    def displayed_dimensional_calibrations(self) -> typing.Sequence[Calibration.Calibration]:
        return self.get_dimensional_calibrations_with_calibration_style(self.dimensional_calibration_style)

    def get_intensity_calibration_with_calibration_style(self, dimensional_calibration_style) -> Calibration.Calibration:
        calibration_style = self.__get_calibration_style_for_id(dimensional_calibration_style)
        if calibration_style is None or self.__data_and_metadata is None:
            if self.__data_and_metadata:
                return self.__data_and_metadata.intensity_calibration
            if self.intensity_calibration:
                return self.intensity_calibration
            return Calibration.Calibration()
        return calibration_style.get_intensity_calibration(self.__data_and_metadata)

    @property
    def displayed_intensity_calibration(self) -> Calibration.Calibration:
        return self.get_intensity_calibration_with_calibration_style(self.dimensional_calibration_style)

    @property
    def calibration_styles(self) -> typing.List[CalibrationStyle]:
        return [CalibrationStyleNative(), CalibrationStylePixelsTopLeft(), CalibrationStylePixelsCenter(),
                CalibrationStyleFractionalTopLeft(), CalibrationStyleFractionalCenter()]

    @property
    def default_calibrated_calibration_style(self):
        return CalibrationStyleNative()

    @property
    def default_uncalibrated_calibration_style(self):
        return CalibrationStylePixelsCenter()

    def __get_calibration_style_for_id(self, calibration_style_id: str) -> typing.Optional[CalibrationStyle]:
        return next(filter(lambda x: x.calibration_style_id == calibration_style_id, self.calibration_styles), None)

    def get_calibration_style_for_id(self, calibration_style_id: str) -> typing.Optional[CalibrationStyle]:
        return next(filter(lambda x: x.calibration_style_id == calibration_style_id, self.calibration_styles), self.default_calibrated_calibration_style)

    def __get_calibrated_value_text(self, value: float, intensity_calibration) -> str:
        if value is not None:
            return intensity_calibration.convert_to_calibrated_value_str(value)
        elif value is None:
            return _("N/A")
        else:
            return str(value)

    def get_value_and_position_text(self, pos) -> (str, str):
        data_and_metadata = self.__data_and_metadata
        dimensional_calibrations = self.displayed_dimensional_calibrations
        intensity_calibration = self.displayed_intensity_calibration

        if data_and_metadata is None or pos is None:
            return str(), str()

        is_sequence = data_and_metadata.is_sequence
        collection_dimension_count = data_and_metadata.collection_dimension_count
        datum_dimension_count = data_and_metadata.datum_dimension_count
        if is_sequence:
            pos = (self.sequence_index, ) + pos
        if collection_dimension_count == 2 and datum_dimension_count == 1:
            pos = pos + (self.slice_center, )
        else:
            pos = tuple(self.collection_index[0:collection_dimension_count]) + pos

        while len(pos) < data_and_metadata.datum_dimension_count:
            pos = (0,) + tuple(pos)

        assert len(pos) == len(data_and_metadata.dimensional_shape)

        position_text = ""
        value_text = ""
        data_shape = data_and_metadata.data_shape
        if len(pos) == 4:
            # 4d image
            # make sure the position is within the bounds of the image
            if 0 <= pos[0] < data_shape[0] and 0 <= pos[1] < data_shape[1] and 0 <= pos[2] < data_shape[2] and 0 <= pos[3] < data_shape[3]:
                position_text = u"{0}, {1}, {2}, {3}".format(
                    dimensional_calibrations[3].convert_to_calibrated_value_str(pos[3], value_range=(0, data_shape[3]), samples=data_shape[3]),
                    dimensional_calibrations[2].convert_to_calibrated_value_str(pos[2], value_range=(0, data_shape[2]), samples=data_shape[2]),
                    dimensional_calibrations[1].convert_to_calibrated_value_str(pos[1], value_range=(0, data_shape[1]), samples=data_shape[1]),
                    dimensional_calibrations[0].convert_to_calibrated_value_str(pos[0], value_range=(0, data_shape[0]), samples=data_shape[0]))
                value_text = self.__get_calibrated_value_text(data_and_metadata.get_data_value(pos), intensity_calibration)
        if len(pos) == 3:
            # 3d image
            # make sure the position is within the bounds of the image
            if 0 <= pos[0] < data_shape[0] and 0 <= pos[1] < data_shape[1] and 0 <= pos[2] < data_shape[2]:
                position_text = u"{0}, {1}, {2}".format(dimensional_calibrations[2].convert_to_calibrated_value_str(pos[2], value_range=(0, data_shape[2]), samples=data_shape[2]),
                    dimensional_calibrations[1].convert_to_calibrated_value_str(pos[1], value_range=(0, data_shape[1]), samples=data_shape[1]),
                    dimensional_calibrations[0].convert_to_calibrated_value_str(pos[0], value_range=(0, data_shape[0]), samples=data_shape[0]))
                value_text = self.__get_calibrated_value_text(data_and_metadata.get_data_value(pos), intensity_calibration)
        if len(pos) == 2:
            # 2d image
            # make sure the position is within the bounds of the image
            if len(data_shape) == 1:
                if pos[-1] >= 0 and pos[-1] < data_shape[-1]:
                    position_text = u"{0}".format(dimensional_calibrations[-1].convert_to_calibrated_value_str(pos[-1], value_range=(0, data_shape[-1]), samples=data_shape[-1]))
                    full_pos = [0, ] * len(data_shape)
                    full_pos[-1] = pos[-1]
                    value_text = self.__get_calibrated_value_text(data_and_metadata.get_data_value(full_pos), intensity_calibration)
            else:
                if pos[0] >= 0 and pos[0] < data_shape[0] and pos[1] >= 0 and pos[1] < data_shape[1]:
                    is_polar = dimensional_calibrations[0].units.startswith("1/") and dimensional_calibrations[0].units == dimensional_calibrations[1].units
                    is_polar = is_polar and abs(dimensional_calibrations[0].scale * data_shape[0] - dimensional_calibrations[1].scale * data_shape[1]) < 1e-12
                    is_polar = is_polar and abs(dimensional_calibrations[0].offset / (dimensional_calibrations[0].scale * data_shape[0]) + 0.5) < 1e-12
                    is_polar = is_polar and abs(dimensional_calibrations[1].offset / (dimensional_calibrations[1].scale * data_shape[1]) + 0.5) < 1e-12
                    if is_polar:
                        x = dimensional_calibrations[1].convert_to_calibrated_value(pos[1])
                        y = dimensional_calibrations[0].convert_to_calibrated_value(pos[0])
                        r = math.sqrt(x * x + y * y)
                        angle = -math.atan2(y, x)
                        r_str = dimensional_calibrations[0].convert_to_calibrated_value_str(dimensional_calibrations[0].convert_from_calibrated_value(r), value_range=(0, data_shape[0]), samples=data_shape[0], display_inverted=True)
                        position_text = u"{0}, {1:.4f}° ({2})".format(r_str, math.degrees(angle), _("polar"))
                    else:
                        position_text = u"{0}, {1}".format(dimensional_calibrations[1].convert_to_calibrated_value_str(pos[1], value_range=(0, data_shape[1]), samples=data_shape[1], display_inverted=True),
                            dimensional_calibrations[0].convert_to_calibrated_value_str(pos[0], value_range=(0, data_shape[0]), samples=data_shape[0], display_inverted=True))
                    value_text = self.__get_calibrated_value_text(data_and_metadata.get_data_value(pos), intensity_calibration)
        if len(pos) == 1:
            # 1d plot
            # make sure the position is within the bounds of the line plot
            if pos[0] >= 0 and pos[0] < data_shape[-1]:
                position_text = u"{0}".format(dimensional_calibrations[-1].convert_to_calibrated_value_str(pos[0], value_range=(0, data_shape[-1]), samples=data_shape[-1]))
                full_pos = [0, ] * len(data_shape)
                full_pos[-1] = pos[0]
                value_text = self.__get_calibrated_value_text(data_and_metadata.get_data_value(full_pos), intensity_calibration)
        return position_text, value_text


def display_factory(lookup_id):
    return Display()



"""
    A module for handle .h5 files for Swift.
"""

import io
import json
import os
import pathlib
import threading

import h5py
import numpy

from nion.swift.model import Utility
from nion.utils import Geometry


def make_directory_if_needed(directory_path):
    """
        Make the directory path, if needed.
    """
    if os.path.exists(directory_path):
        if not os.path.isdir(directory_path):
            raise OSError("Path is not a directory:", directory_path)
    else:
        os.makedirs(directory_path)


class HDF5Handler:

    def __init__(self, file_path):
        self.__file_path = str(file_path)
        self.__lock = threading.RLock()
        self.__fp = None
        self.__dataset = None

    def close(self):
        if self.__fp:
            self.__fp.close()
            self.__fp = None

    @property
    def reference(self):
        return self.__file_path

    @property
    def is_valid(self):
        return True

    @classmethod
    def is_matching(self, file_path):
        if file_path.endswith(".h5") and os.path.exists(file_path):
            return True
        return False

    @classmethod
    def make(cls, file_path):
        return cls(cls.make_path(file_path))

    @classmethod
    def make_path(cls, file_path) -> str:
        return str(pathlib.Path(file_path).with_suffix(cls.get_extension()))

    @classmethod
    def get_extension(self) -> str:
        return ".h5"

    def __ensure_open(self):
        if not self.__fp:
            make_directory_if_needed(os.path.dirname(self.__file_path))
            self.__fp = h5py.File(self.__file_path)

    def __write_properties_to_dataset(self, properties):
        with self.__lock:
            assert self.__dataset is not None

            class JSONEncoder(json.JSONEncoder):
                def default(self, obj):
                    if isinstance(obj, Geometry.IntPoint) or isinstance(obj, Geometry.IntSize) or isinstance(obj, Geometry.IntRect) or isinstance(obj, Geometry.FloatPoint) or isinstance(obj, Geometry.FloatSize) or isinstance(obj, Geometry.FloatRect):
                        return tuple(obj)
                    else:
                        return json.JSONEncoder.default(self, obj)

            json_io = io.StringIO()
            json.dump(Utility.clean_dict(properties), json_io, cls=JSONEncoder)
            json_str = json_io.getvalue()

            self.__dataset.attrs["properties"] = json_str

    def __ensure_dataset(self):
        with self.__lock:
            self.__ensure_open()
            if "data" in self.__fp:
                self.__dataset = self.__fp["data"]
            else:
                self.__dataset = self.__fp.create_dataset("data", data=numpy.empty((0,)))

    def write_data(self, data, file_datetime):
        with self.__lock:
            assert data is not None
            self.__ensure_open()
            json_properties = None
            # handle three cases:
            #   1 - 'data' doesn't yet exist (require_dataset)
            #   2 - 'data' exists but is a different size (delete, then require_dataset)
            #   3 - 'data' exists and is the same size (overwrite)
            if not "data" in self.__fp:
                # case 1
                self.__dataset = self.__fp.require_dataset("data", shape=data.shape, dtype=data.dtype, data=data)
            else:
                self.__dataset = self.__fp["data"]
                if self.__dataset.shape != data.shape or self.__dataset.dtype != data.dtype:
                    # case 2
                    json_properties = self.__dataset.attrs.get("properties", "")
                    self.__dataset = None
                    self.__fp.close()
                    self.__fp = None
                    os.remove(self.__file_path)
                    self.__ensure_open()
                    self.__dataset = self.__fp.require_dataset("data", shape=data.shape, dtype=data.dtype, data=data)
                else:
                    # case 3
                    self.__dataset[:] = data
            if json_properties is not None:
                self.__dataset.attrs["properties"] = json_properties
            self.__fp.flush()

    def write_properties(self, properties, file_datetime):
        with self.__lock:
            self.__ensure_open()
            self.__ensure_dataset()
            self.__write_properties_to_dataset(properties)
            self.__fp.flush()

    def read_properties(self):
        with self.__lock:
            self.__ensure_open()
            self.__ensure_dataset()
            json_properties = self.__dataset.attrs.get("properties", "")
            return json.loads(json_properties)

    def read_data(self):
        with self.__lock:
            self.__ensure_open()
            self.__ensure_dataset()
            if self.__dataset.shape == (0, ):
                return None
            return self.__dataset

    def remove(self):
        self.close()
        if os.path.isfile(self.__file_path):
            os.remove(self.__file_path)



import typing

# standardized metadata paths, mapping to properties

session_key_map = {
    'stem.session.site': {'path': ['site'], 'type': 'string'},
    'stem.session.instrument': {'path': ['instrument'], 'type': 'string'},
    'stem.session.detector': {'path': ['detector'], 'type': 'string'},
    'stem.session.task': {'path': ['task'], 'type': 'string'},
    'stem.session.microscopist': {'path': ['microscopist'], 'type': 'string'},
    'stem.session.sample': {'path': ['sample'], 'type': 'string'},
    'stem.session.sample_area': {'path': ['sample_area'], 'type': 'string'},
    'stem.session.sample_source': {'path': ['sample_source'], 'type': 'string'},
    'stem.session.sample_formula': {'path': ['sample_formula'], 'type': 'string'},
}

# 'hardware_source' should be 'detector' at some point in the future. see HMSA file format for more thoughts.

# keys can exclude the unit suffix if units are SI

key_map = {
    'stem.hardware_source.id': {'path': ['hardware_source', 'hardware_source_id'], 'type': 'string'},
    'stem.hardware_source.name': {'path': ['hardware_source', 'hardware_source_name'], 'type': 'string'},

    'stem.signal_type': {'path': ['hardware_source', 'signal_type'], 'type': 'string'},
    # EDS, WDS, ELS, AES, PES, XRF, CLS, GAM, BEI, CBED, EBSD, EDIF, LEED, OPR, OPT, PIXE, RHEED, SEI, SXES, TEM
    # see HMSA

    'stem.high_tension_v': {'path': ['hardware_source', 'autostem', 'high_tension_v'], 'type': 'integer'},
    'stem.gun_type': {'path': ['hardware_source', 'gun_type'], 'type': 'string'},
    'stem.convergence_angle_rad': {'path': ['hardware_source', 'convergence_angle_rad'], 'type': 'real'},
    'stem.collection_angle_rad': {'path': ['hardware_source', 'collection_angle_rad'], 'type': 'real'},
    'stem.probe_size_m2': {'path': ['hardware_source', 'probe_size_m2'], 'type': 'real'},
    'stem.beam_current_a': {'path': ['hardware_source', 'beam_current_a'], 'type': 'real'},
    'stem.defocus_m': {'path': ['hardware_source', 'defocus_m'], 'type': 'real'},

    'stem.eels.spectrum_type': {'path': ['hardware_source', 'eels_spectrum_type'], 'type': 'string'},
    'stem.eels.resolution_eV': {'path': ['hardware_source', 'eels_resolution_eV'], 'type': 'string'},
    'stem.eels.is_monochromated': {'path': ['hardware_source', 'eels_is_monochromated'], 'type': 'boolean'},

    'stem.camera.binning': {'path': ['hardware_source', 'binning'], 'type': 'integer'},
    'stem.camera.channel_id': {'path': ['hardware_source', 'channel_id'], 'type': 'string'},
    'stem.camera.channel_index': {'path': ['hardware_source', 'channel_index'], 'type': 'integer'},
    'stem.camera.channel_name': {'path': ['hardware_source', 'channel_name'], 'type': 'string'},
    'stem.camera.exposure_s': {'path': ['hardware_source', 'exposure'], 'type': 'real'},
    'stem.camera.frame_index': {'path': ['hardware_source', 'frame_index'], 'type': 'integer'},
    'stem.camera.valid_rows': {'path': ['hardware_source', 'valid_rows'], 'type': 'integer'},
    'stem.camera.detector_current': {'path': ['hardware_source', 'detector_current'], 'type': 'real'},

    'stem.scan.center_x_nm': {'path': ['hardware_source', 'center_x_nm'], 'type': 'real'},
    'stem.scan.center_y_nm': {'path': ['hardware_source', 'center_y_nm'], 'type': 'real'},
    'stem.scan.channel_id': {'path': ['hardware_source', 'channel_id'], 'type': 'string'},
    'stem.scan.channel_index': {'path': ['hardware_source', 'channel_index'], 'type': 'integer'},
    'stem.scan.channel_name': {'path': ['hardware_source', 'channel_name'], 'type': 'string'},
    'stem.scan.frame_time_s': {'path': ['hardware_source', 'exposure'], 'type': 'real'},
    'stem.scan.fov_nm': {'path': ['hardware_source', 'fov_nm'], 'type': 'real'},
    'stem.scan.frame_index': {'path': ['hardware_source', 'frame_index'], 'type': 'integer'},
    'stem.scan.pixel_time_us': {'path': ['hardware_source', 'pixel_time_us'], 'type': 'real'},
    'stem.scan.rotation_rad': {'path': ['hardware_source', 'rotation_rad'], 'type': 'real'},
    'stem.scan.scan_id': {'path': ['hardware_source', 'scan_id'], 'type': 'string'},
    'stem.scan.valid_rows': {'path': ['hardware_source', 'valid_rows'], 'type': 'integer'},
    'stem.scan.line_time_us': {'path': ['hardware_source', 'line_time_us'], 'type': 'real'},
}


def has_metadata_value(metadata_source, key: str) -> bool:
    """Return whether the metadata value for the given key exists.

    There are a set of predefined keys that, when used, will be type checked and be interoperable with other
    applications. Please consult reference documentation for valid keys.

    If using a custom key, we recommend structuring your keys in the '<group>.<attribute>' format followed
    by the predefined keys. e.g. 'session.instrument' or 'camera.binning'.

    Also note that some predefined keys map to the metadata ``dict`` but others do not. For this reason, prefer
    using the ``metadata_value`` methods over directly accessing ``metadata``.
    """
    desc = session_key_map.get(key)
    if desc is not None:
        d = getattr(metadata_source, "session_metadata", dict())
        for k in desc['path'][:-1]:
            d =  d.setdefault(k, dict()) if d is not None else None
        if d is not None:
            return desc['path'][-1] in d
    desc = key_map.get(key)
    if desc is not None:
        d = getattr(metadata_source, "metadata", dict())
        for k in desc['path'][:-1]:
            d =  d.setdefault(k, dict()) if d is not None else None
        if d is not None:
            return desc['path'][-1] in d
    raise False

def get_metadata_value(metadata_source, key: str) -> typing.Any:
    """Get the metadata value for the given key.

    There are a set of predefined keys that, when used, will be type checked and be interoperable with other
    applications. Please consult reference documentation for valid keys.

    If using a custom key, we recommend structuring your keys in the '<group>.<attribute>' format followed
    by the predefined keys. e.g. 'session.instrument' or 'camera.binning'.

    Also note that some predefined keys map to the metadata ``dict`` but others do not. For this reason, prefer
    using the ``metadata_value`` methods over directly accessing ``metadata``.
    """
    desc = session_key_map.get(key)
    if desc is not None:
        v = getattr(metadata_source, "session_metadata", dict())
        for k in desc['path']:
            v =  v.get(k) if v is not None else None
        return v
    desc = key_map.get(key)
    if desc is not None:
        v = getattr(metadata_source, "metadata", dict())
        for k in desc['path']:
            v =  v.get(k) if v is not None else None
        return v
    raise KeyError()

def set_metadata_value(metadata_source, key: str, value: typing.Any) -> None:
    """Set the metadata value for the given key.

    There are a set of predefined keys that, when used, will be type checked and be interoperable with other
    applications. Please consult reference documentation for valid keys.

    If using a custom key, we recommend structuring your keys in the '<group>.<attribute>' format followed
    by the predefined keys. e.g. 'session.instrument' or 'camera.binning'.

    Also note that some predefined keys map to the metadata ``dict`` but others do not. For this reason, prefer
    using the ``metadata_value`` methods over directly accessing ``metadata``.
    """
    desc = session_key_map.get(key)
    if desc is not None:
        d0 = getattr(metadata_source, "session_metadata", dict())
        d = d0
        for k in desc['path'][:-1]:
            d =  d.setdefault(k, dict()) if d is not None else None
        if d is not None:
            d[desc['path'][-1]] = value
            metadata_source.session_metadata = d0
            return
    desc = key_map.get(key)
    if desc is not None:
        d0 = getattr(metadata_source, "metadata", dict())
        d = d0
        for k in desc['path'][:-1]:
            d =  d.setdefault(k, dict()) if d is not None else None
        if d is not None:
            d[desc['path'][-1]] = value
            metadata_source.metadata = d0
            return
    raise KeyError()

def delete_metadata_value(metadata_source, key: str) -> None:
    """Delete the metadata value for the given key.

    There are a set of predefined keys that, when used, will be type checked and be interoperable with other
    applications. Please consult reference documentation for valid keys.

    If using a custom key, we recommend structuring your keys in the '<dotted>.<group>.<attribute>' format followed
    by the predefined keys. e.g. 'stem.session.instrument' or 'stm.camera.binning'.

    Also note that some predefined keys map to the metadata ``dict`` but others do not. For this reason, prefer
    using the ``metadata_value`` methods over directly accessing ``metadata``.
    """
    desc = session_key_map.get(key)
    if desc is not None:
        d0 = getattr(metadata_source, "session_metadata", dict())
        d = d0
        for k in desc['path'][:-1]:
            d =  d.setdefault(k, dict()) if d is not None else None
        if d is not None and desc['path'][-1] in d:
            d.pop(desc['path'][-1], None)
            metadata_source.session_metadata = d0
            return
    desc = key_map.get(key)
    if desc is not None:
        d0 = getattr(metadata_source, "metadata", dict())
        d = d0
        for k in desc['path'][:-1]:
            d =  d.setdefault(k, dict()) if d is not None else None
        if d is not None and desc['path'][-1] in d:
            d.pop(desc['path'][-1], None)
            metadata_source.metadata = d0
            return



# standard libraries
import copy
import gettext
import logging
import threading
import time

# third party libraries
# None

# local libraries
from nion.swift import Panel
from nion.swift.model import Utility
from nion.utils import Event
from nion.utils import Observable

_ = gettext.gettext


"""
    Task modules can register task viewer factory with the task manager.

    Tasks will request a new task controller from the document model.

    A task section will get created and added to the task panel with some type of progress
    indicator and cancel button. The task section will display a task viewer if one is
    available for the task data.

    While the task is running, it is free to call task controller methods from
    a thread to update the data. The task controller will take care of getting the
    data to the task viewer on the UI thread.

    When the task finishes, it can optionally send final data. The last data available
    data will get permanently stored into the document.

    When the document loads, all tasks will be displayed in the task panel. The
    user has an option to copy, export, or delete finished tasks.

"""

class TaskPanel(Panel.Panel):

    def __init__(self, document_controller, panel_id, properties):
        super().__init__(document_controller, panel_id, _("Tasks"))

        # thread safe
        def task_created(task):
            with self.__pending_tasks_mutex:
                self.__pending_tasks.append(task)
                self.document_controller.add_task(str(id(self)), self.__perform_tasks)

        # connect to the document controller
        self.__task_created_event_listener = self.document_controller.task_created_event.listen(task_created)

        # the main column widget contains a stack group for each operation
        self.column = self.ui.create_column_widget(properties=properties)  # TODO: put this in scroll area
        self.scroll_area = self.ui.create_scroll_area_widget()
        self.scroll_area.set_scrollbar_policies("off", "needed")
        self.task_column_container = self.ui.create_column_widget()
        self.task_column = self.ui.create_column_widget()
        self.task_column_container.add(self.task_column)
        self.task_column_container.add_stretch()
        self.scroll_area.content = self.task_column_container
        self.column.add(self.scroll_area)
        self.widget = self.column

        # map task to widgets
        self.__pending_tasks = list()
        self.__pending_tasks_mutex = threading.RLock()
        self.__task_needs_update = set()
        self.__task_needs_update_mutex = threading.RLock()
        self.__task_section_controller_list = list()
        self.__task_changed_event_listeners = list()

    def close(self):
        self.document_controller.clear_task(str(id(self)))
        # disconnect to the document controller
        self.__task_created_event_listener.close()
        self.__task_created_event_listener = None
        # disconnect from tasks
        for task_changed_event_listener in self.__task_changed_event_listeners:
            task_changed_event_listener.close()
        self.__task_changed_event_listeners = None
        # finish closing
        super().close()

    # not thread safe
    def __perform_tasks(self):
        # for all pending tasks, make a task section controller, then add the task
        # to the needs update list.
        with self.__pending_tasks_mutex:
            pending_tasks = self.__pending_tasks
            self.__pending_tasks = list()
        for task in pending_tasks:
            task_section_controller = TaskSectionController(self.ui, task)
            self.task_column.insert(task_section_controller.widget, 0)
            self.scroll_area.scroll_to(0, 0)
            self.__task_section_controller_list.append(task_section_controller)
            with self.__task_needs_update_mutex:
                self.__task_needs_update.add(task)

            # thread safe
            def task_changed():
                with self.__task_needs_update_mutex:
                    self.__task_needs_update.add(task)
                    self.document_controller.add_task(str(id(self)), self.__perform_tasks)

            # TODO: currently, tasks don't get deleted since they are displayed until exit.
            # add the listener to a never deleted list.
            self.__task_changed_event_listeners.append(task.task_changed_event.listen(task_changed))

        # for each started task (i.e. has a section controller), if it needs an
        # update, update it in the section controller.
        for task_section_controller in self.__task_section_controller_list:
            task = task_section_controller.task
            if task in self.__task_needs_update:
                # remove from update list before updating to prevent a race
                with self.__task_needs_update_mutex:
                    self.__task_needs_update.remove(task)
                # update
                task_section_controller.update()
        # finally, if there are tasks that need updating, run perform task again.
        if len(self.__task_needs_update) > 0:
            self.document_controller.add_task(str(id(self)), self.__perform_tasks)


# this is UI object, and not a thread safe
class TaskSectionController(object):

    def __init__(self, ui, task):
        self.ui = ui
        self.task = task

        widget = self.ui.create_column_widget()
        task_header = self.ui.create_row_widget()
        self.title_widget = self.ui.create_label_widget(properties={"stylesheet": "font-weight: bold"})
        task_header.add(self.title_widget)
        task_spacer_row = self.ui.create_row_widget()
        task_spacer_row_col = self.ui.create_column_widget()
        task_spacer_row.add_spacing(20)
        task_spacer_row.add(task_spacer_row_col)
        self.task_progress_row = self.ui.create_row_widget()
        self.task_progress_label = self.ui.create_label_widget()
        self.task_progress_row.add(self.task_progress_label)
        task_time_row = self.ui.create_row_widget()
        self.task_progress_state = self.ui.create_label_widget(properties={"stylesheet": "font: italic"})
        task_time_row.add(self.task_progress_state)
        task_spacer_row_col.add(self.task_progress_row)
        task_spacer_row_col.add(task_time_row)

        # add custom ui, if any
        self.task_ui_controller = TaskManager().build_task_ui(self.ui, task)
        if self.task_ui_controller:
            task_spacer_row_col.add(self.task_ui_controller.widget)

        widget.add(task_header)
        widget.add(task_spacer_row)

        self.__title_text = None
        copy_button = self.ui.create_push_button_widget(_("Copy to Clipboard"))
        def copy_to_clipboard():
            clipboard_text = self.task_ui_controller.clipboard_text
            if clipboard_text:
                if self.__title_text:
                    clipboard_text = self.__title_text + "\n" + clipboard_text
                self.ui.clipboard_set_text(clipboard_text + "\n")
        copy_button.on_clicked = copy_to_clipboard
        copy_button_row = self.ui.create_row_widget()
        copy_button_row.add(copy_button)
        copy_button_row.add_stretch()
        widget.add(copy_button_row)

        self.widget = widget

        self.update()

    # only called on UI thread
    def update(self):

        # update the title
        self.__title_text = str(self.task.title).upper()
        self.title_widget.text = str(self.task.title)

        # update the progress label
        in_progress = self.task.in_progress
        if in_progress:
            self.task_progress_label.visible = True
            done_percentage_str = "{0:.0f}%".format(float(self.task.progress[0])/self.task.progress[1] * 100) if self.task.progress else "--"
            progress_text = "{0} {1}".format(done_percentage_str, self.task.progress_text)
            self.task_progress_label.text = progress_text
            self.__title_text += " (" + progress_text + ")"
        else:
            self.task_progress_label.visible = False

        # update the state text
        task_state_str = _("In Progress") if in_progress else _("Done")
        task_time_str = time.strftime("%c", time.localtime(self.task.start_time if in_progress else self.task.finish_time))
        progress_state_text = "{} {}".format(task_state_str, task_time_str)
        self.task_progress_state.text = progress_state_text
        self.__title_text += "\n" + progress_state_text

        # update the custom builder, if any
        if self.task_ui_controller:
            self.task_ui_controller.update_task(self.task)


class Task(Observable.Observable):

    def __init__(self, title, task_type):
        super().__init__()
        self.__title = title
        self.__start_time = None
        self.__finish_time = None
        self.__task_type = task_type
        self.__task_data = None
        self.__task_data_mutex = threading.RLock()
        self.__progress = None
        self.__progress_text = str()
        self.task_changed_event = Event.Event()

    # title
    @property
    def title(self):
        return self.__title

    @title.setter
    def title(self, value):
        self.__title = value
        self.notify_property_changed("title")
        self.task_changed_event.fire()

    # start time
    @property
    def start_time(self):
        return self.__start_time

    @start_time.setter
    def start_time(self, value):
        self.__start_time = value
        self.notify_property_changed("start_time")
        self.task_changed_event.fire()

    # finish time
    @property
    def finish_time(self):
        return self.__finish_time

    @finish_time.setter
    def finish_time(self, value):
        self.__finish_time = value
        self.notify_property_changed("finish_time")
        self.task_changed_event.fire()

    # in progress
    @property
    def in_progress(self):
        return self.finish_time is None

    # progress
    @property
    def progress(self):
        return self.__progress

    @progress.setter
    def progress(self, value):
        self.__progress = value
        self.task_changed_event.fire()

    # progress_text
    @property
    def progress_text(self):
        return self.__progress_text

    @progress_text.setter
    def progress_text(self, value):
        self.__progress_text = value
        self.task_changed_event.fire()

    # task type
    @property
    def task_type(self):
        return self.__task_type

    # task data
    @property
    def task_data(self):
        with self.__task_data_mutex:
            return copy.copy(self.__task_data)

    @task_data.setter
    def task_data(self, task_data):
        with self.__task_data_mutex:
            self.__task_data = copy.copy(task_data)
        self.notify_property_changed("task_data")
        self.task_changed_event.fire()


# all public methods are thread safe
class TaskContextManager(object):

    def __init__(self, container, task, logging):
        self.__container = container
        self.__task = task
        self.__logging = logging

    def __enter__(self):
        if self.__logging:
            logging.debug("%s: started", self.__task.title)
        self.__task.start_time = time.time()
        return self

    def __exit__(self, type, value, traceback):
        self.__task.finish_time = time.time()
        if self.__logging:
            logging.debug("%s: finished", self.__task.title)

    def update_progress(self, progress_text, progress=None, task_data=None):
        self.__task.progress_text = progress_text
        self.__task.progress = progress
        if task_data:
            self.__task.task_data = task_data
        if self.__logging:
            logging.debug("%s: %s %s", self.__task.title, progress_text, progress if progress else "")


class TaskManager(metaclass=Utility.Singleton):

    def __init__(self):
        self.__task_ui_builder_map = dict()

    def register_task_type_builder(self, task_type, fn):
        self.__task_ui_builder_map[task_type] = fn

    def unregister_task_type_builder(self, task_type):
        del self.__task_ui_builder_map[task_type]

    def build_task_ui(self, ui, task):
        if task.task_type in self.__task_ui_builder_map:
            return self.__task_ui_builder_map[task.task_type](ui)
        return None


class TableController(object):

    def __init__(self, ui):
        self.ui = ui
        self.clipboard_text = None
        self.widget = self.ui.create_column_widget()
        self.column_widgets = self.ui.create_row_widget()
        self.widget.add(self.column_widgets)

    def update_task(self, task):
        if task.task_data:
            column_count = len(task.task_data["headers"])
            while self.column_widgets.child_count > column_count:
                self.column_widgets.remove(self.column_widgets.child_count - 1)
            while self.column_widgets.child_count < column_count:
                self.column_widgets.add(self.ui.create_column_widget())
            row_count = len(task.task_data["data"]) if "data" in task.task_data else 0
            text_lines = [list() for _ in range(row_count + 1)]
            for column_index, column_widget in enumerate(self.column_widgets.children):
                while column_widget.child_count > row_count + 1:
                    column_widget.remove(column_widget.child_count - 1)
                while column_widget.child_count < row_count + 1:
                    properties = {"stylesheet": "font-weight: bold"} if column_widget.child_count == 0 else None
                    column_widget.add(self.ui.create_label_widget(properties=properties))
                header_text = task.task_data["headers"][column_index]
                column_widget.children[0].text = header_text
                text_lines[0].append(header_text)
                for row_index in range(row_count):
                    data_text = str(task.task_data["data"][row_index][column_index])
                    column_widget.children[row_index + 1].text = data_text
                    text_lines[row_index + 1].append(data_text)
            self.clipboard_text = "\n".join(["  ".join(text_line) for text_line in text_lines])
        else:
            self.column_widgets.remove_all()


class StringListController(object):

    def __init__(self, ui):
        self.ui = ui
        self.widget = self.ui.create_label_widget("[]")
        self.clipboard_text = None

    def update_task(self, task):
        strings = task.task_data["strings"] if task.task_data else list()
        self.clipboard_text = "[" + ":".join(strings) + "]"
        self.widget.text = self.clipboard_text

TaskManager().register_task_type_builder("string_list", lambda ui: StringListController(ui))
TaskManager().register_task_type_builder("table", lambda ui: TableController(ui))



# standard libraries
import contextlib
import unittest

# third party libraries
import numpy

# local libraries
from nion.data import Calibration
from nion.data import DataAndMetadata
from nion.swift import Application
from nion.swift import Facade
from nion.swift.model import DocumentModel
from nion.swift.model import DataItem
from nion.swift.model import Graphics
from nion.ui import TestUI
from nion.utils import Geometry


Facade.initialize()


class TestFacadeClass(unittest.TestCase):

    def setUp(self):
        self.app = Application.Application(TestUI.UserInterface(), set_global=True)
        self.app.workspace_dir = str()

    def tearDown(self):
        pass

    def test_basic_api_methods(self):
        memory_persistent_storage_system = DocumentModel.MemoryStorageSystem()
        document_model = DocumentModel.DocumentModel(persistent_storage_system=memory_persistent_storage_system)
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            api = Facade.get_api("~1.0", "~1.0")
            self.assertIsNotNone(api.library)
            self.assertIsNotNone(api.application)
            self.assertIsNotNone(api.create_calibration(1.0, 2.0, "mm"))

    def test_create_data_item_from_data(self):
        memory_persistent_storage_system = DocumentModel.MemoryStorageSystem()
        document_model = DocumentModel.DocumentModel(persistent_storage_system=memory_persistent_storage_system)
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            data0 = numpy.arange(64).reshape(8, 8)
            data_item = DataItem.DataItem(data0)
            document_model.append_data_item(data_item)
            api = Facade.get_api("~1.0", "~1.0")
            library = api.library
            self.assertEqual(library.data_item_count, 1)
            self.assertEqual(len(library.data_items), 1)
            data1 = numpy.arange(128).reshape(16, 8)
            data2 = numpy.arange(128).reshape(8, 16)
            data3 = numpy.arange(16).reshape(4, 4)
            data_item1_ref = library.create_data_item("one")
            with library.data_ref_for_data_item(data_item1_ref) as data_ref:
                data_ref.data = data1
            data_item2_ref = library.create_data_item_from_data(data2, "two")
            data_and_metadata =  api.create_data_and_metadata(data3)
            data_item3_ref = library.create_data_item_from_data_and_metadata(data_and_metadata, "three")
            self.assertEqual(library.data_item_count, 4)
            self.assertTrue(numpy.array_equal(document_model.data_items[1].data, data1))
            self.assertTrue(numpy.array_equal(document_model.data_items[2].data, data2))
            self.assertTrue(numpy.array_equal(document_model.data_items[3].data, data3))

    def test_library_and_data_items_can_be_compared_for_equality(self):
        memory_persistent_storage_system = DocumentModel.MemoryStorageSystem()
        document_model = DocumentModel.DocumentModel(persistent_storage_system=memory_persistent_storage_system)
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            data_item1 = DataItem.DataItem(numpy.zeros((2, 2)))
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.zeros((2, 2)))
            document_model.append_data_item(data_item2)
            api = Facade.get_api("~1.0", "~1.0")
            self.assertEqual(api.library, api.library)
            self.assertEqual(api.library.data_items, api.library.data_items)

    def test_graphic_is_invalid_if_source_is_removed(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            api = Facade.get_api("~1.0", "~1.0")
            library = api.library
            data_item = library.create_data_item_from_data(numpy.zeros((16, 16)))
            self.assertEqual(len(Facade.Graphic.instances), 0)
            graphic = data_item.add_point_region(10, 10)
            self.assertEqual(len(Facade.Graphic.instances), 1)
            graphic = None
            self.assertEqual(len(Facade.Graphic.instances), 0)

    def test_create_data_item_from_data_as_sequence(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            api = Facade.get_api("~1.0", "~1.0")
            library = api.library
            data_and_metadata =  DataAndMetadata.new_data_and_metadata(numpy.zeros((8, 4, 5)), data_descriptor=DataAndMetadata.DataDescriptor(True, 0, 2))
            data_item = library.create_data_item_from_data_and_metadata(data_and_metadata, "three")
            self.assertEqual(library.data_item_count, 1)
            self.assertTrue(document_model.data_items[0].is_sequence)

    def test_data_on_empty_data_item_returns_none(self):
        memory_persistent_storage_system = DocumentModel.MemoryStorageSystem()
        document_model = DocumentModel.DocumentModel(persistent_storage_system=memory_persistent_storage_system)
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            api = Facade.get_api("~1.0", "~1.0")
            library = api.library
            data_item1_ref = library.create_data_item("one")
            with library.data_ref_for_data_item(data_item1_ref) as data_ref:
                self.assertIsNone(data_ref.data)

    def test_data_item_data_methods(self):
        memory_persistent_storage_system = DocumentModel.MemoryStorageSystem()
        document_model = DocumentModel.DocumentModel(persistent_storage_system=memory_persistent_storage_system)
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            data0 = numpy.arange(64).reshape(8, 8)
            data_item = DataItem.DataItem(data0)
            document_model.append_data_item(data_item)
            api = Facade.get_api("~1.0", "~1.0")
            library = api.library
            data1 = numpy.arange(128).reshape(16, 8)
            data_item_ref = library.data_items[0]
            self.assertTrue(numpy.array_equal(data_item_ref.data, data0))
            data_item_ref.set_data(data1)
            self.assertTrue(numpy.array_equal(data_item_ref.data, data1))
            data2 = numpy.arange(128).reshape(8, 16)
            data_item_ref.set_data_and_metadata(api.create_data_and_metadata(data2))
            self.assertTrue(numpy.array_equal(data_item_ref.data, data2))
            self.assertTrue(numpy.array_equal(data_item_ref.data_and_metadata.data, data2))

    def test_data_item_metadata_methods(self):
        memory_persistent_storage_system = DocumentModel.MemoryStorageSystem()
        document_model = DocumentModel.DocumentModel(persistent_storage_system=memory_persistent_storage_system)
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            data0 = numpy.arange(64).reshape(8, 8)
            data_item = DataItem.DataItem(data0)
            data_item.set_intensity_calibration(Calibration.Calibration(0.1, 0.2, "dogs"))
            data_item.set_dimensional_calibrations([Calibration.Calibration(0.3, 0.4, "cats"), Calibration.Calibration(0.5, 0.6, "cats")])
            metadata = {"title": "Dogs eat cats."}
            data_item.metadata = metadata
            document_model.append_data_item(data_item)
            api = Facade.get_api("~1.0", "~1.0")
            library = api.library
            data_item_ref = library.data_items[0]
            self.assertEqual(data_item_ref.intensity_calibration.units, "dogs")
            self.assertEqual(data_item_ref.dimensional_calibrations[1].units, "cats")
            self.assertEqual(data_item_ref.metadata, metadata)
            data_item_ref.set_intensity_calibration(api.create_calibration(0.11, 0.22, "cats"))
            data_item_ref.set_dimensional_calibrations([api.create_calibration(0.33, 0.44, "mice"), api.create_calibration(0.44, 0.66, "mice")])
            metadata2 = {"title": "Cats eat mice."}
            data_item_ref.set_metadata(metadata2)
            self.assertAlmostEqual(data_item.intensity_calibration.offset, 0.11)
            self.assertAlmostEqual(data_item.dimensional_calibrations[0].offset, 0.33)
            self.assertEqual(data_item.metadata, metadata2)

    def test_data_item_regions(self):
        memory_persistent_storage_system = DocumentModel.MemoryStorageSystem()
        document_model = DocumentModel.DocumentModel(persistent_storage_system=memory_persistent_storage_system)
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.arange(64).reshape(8, 8))
            document_model.append_data_item(data_item)
            data_item_1d = DataItem.DataItem(numpy.arange(32).reshape(32))
            document_model.append_data_item(data_item_1d)
            api = Facade.get_api("~1.0", "~1.0")
            library = api.library
            data_item_ref = library.data_items[0]
            data_item_1d_ref = library.data_items[1]
            r0 = data_item_ref.add_point_region(0.1, 0.2)
            r1 = data_item_ref.add_rectangle_region(0.3, 0.4, 0.5, 0.6)
            r2 = data_item_ref.add_ellipse_region(0.3, 0.4, 0.5, 0.6)
            r3 = data_item_ref.add_line_region(0.1, 0.2, 0.3, 0.4)
            r4 = data_item_1d_ref.add_interval_region(0.1, 0.2)
            r5 = data_item_1d_ref.add_channel_region(0.5)
            r0.label = "One"
            self.assertEqual(r0.type, "point-region")
            self.assertEqual(r1.type, "rectangle-region")
            self.assertEqual(r2.type, "ellipse-region")
            self.assertEqual(r3.type, "line-region")
            self.assertEqual(r4.type, "interval-region")
            self.assertEqual(r5.type, "channel-region")
            r4.set_property("end", 0.3)
            self.assertAlmostEqual(r4.get_property("end"), 0.3)
            self.assertEqual(len(data_item.displays[0].graphics), 4)
            self.assertEqual(len(data_item_1d.displays[0].graphics), 2)
            self.assertIsInstance(data_item.displays[0].graphics[0], Graphics.PointGraphic)
            self.assertIsInstance(data_item.displays[0].graphics[1], Graphics.RectangleGraphic)
            self.assertIsInstance(data_item.displays[0].graphics[2], Graphics.EllipseGraphic)
            self.assertIsInstance(data_item.displays[0].graphics[3], Graphics.LineGraphic)
            self.assertIsInstance(data_item_1d.displays[0].graphics[0], Graphics.IntervalGraphic)
            self.assertIsInstance(data_item_1d.displays[0].graphics[1], Graphics.ChannelGraphic)

    def test_display_data_panel_reuses_existing_display(self):
        memory_persistent_storage_system = DocumentModel.MemoryStorageSystem()
        document_model = DocumentModel.DocumentModel(persistent_storage_system=memory_persistent_storage_system)
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            # configure data item
            data_item = DataItem.DataItem(numpy.arange(64).reshape(8, 8))
            document_model.append_data_item(data_item)
            # configure workspace
            d = {"type": "splitter", "orientation": "vertical", "splits": [0.5, 0.5], "children": [
                {"type": "image", "uuid": "0569ca31-afd7-48bd-ad54-5e2bb9f21102", "identifier": "a", "selected": True},
                {"type": "image", "uuid": "acd77f9f-2f6f-4fbf-af5e-94330b73b997", "identifier": "b"}]}
            workspace_2x1 = document_controller.workspace_controller.new_workspace("2x1", d)
            document_controller.workspace_controller.change_workspace(workspace_2x1)
            root_canvas_item = document_controller.workspace_controller.image_row.children[0]._root_canvas_item()
            root_canvas_item.layout_immediate(Geometry.IntSize(width=640, height=480))
            self.assertIsNone(document_controller.workspace_controller.display_panels[0].data_item)
            self.assertIsNone(document_controller.workspace_controller.display_panels[1].data_item)
            # test display_data_item
            api = Facade.get_api("~1.0", "~1.0")
            library = api.library
            document_controller_ref = api.application.document_controllers[0]
            data_item_ref = library.data_items[0]
            # display data item and verify it is displayed
            display_panal_ref = document_controller_ref.display_data_item(data_item_ref)
            self.assertEqual(document_controller.workspace_controller.display_panels[0].data_item, data_item_ref._data_item)
            self.assertIsNone(document_controller.workspace_controller.display_panels[1].data_item)
            self.assertEqual(document_controller.workspace_controller.display_panels[0], display_panal_ref._display_panel)
            # display data item again and verify it is displayed only once
            display_panal_ref = document_controller_ref.display_data_item(data_item_ref)
            self.assertEqual(document_controller.workspace_controller.display_panels[0].data_item, data_item_ref._data_item)
            self.assertIsNone(document_controller.workspace_controller.display_panels[1].data_item)
            self.assertEqual(document_controller.workspace_controller.display_panels[0], display_panal_ref._display_panel)

    def test_target_data_item_returns_none_if_panel_is_empty(self):
        memory_persistent_storage_system = DocumentModel.MemoryStorageSystem()
        document_model = DocumentModel.DocumentModel(persistent_storage_system=memory_persistent_storage_system)
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            # configure workspace
            workspace_1x1 = document_controller.document_model.workspaces[0]
            document_controller.workspace_controller.change_workspace(workspace_1x1)
            display_panel = document_controller.selected_display_panel

            api = Facade.get_api("~1.0", "~1.0")
            # the display is already filled. display panel should be None.
            self.assertIsNone(api.application.document_windows[0].target_data_item)

    def test_display_data_item_returns_none_if_no_panel_available(self):
        memory_persistent_storage_system = DocumentModel.MemoryStorageSystem()
        document_model = DocumentModel.DocumentModel(persistent_storage_system=memory_persistent_storage_system)
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            # configure data item
            data_item1 = DataItem.DataItem(numpy.zeros((8, 8)))
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.zeros((8, 8)))
            document_model.append_data_item(data_item2)
            # configure workspace
            workspace_1x1 = document_controller.document_model.workspaces[0]
            document_controller.workspace_controller.change_workspace(workspace_1x1)
            display_panel = document_controller.selected_display_panel

            api = Facade.get_api("~1.0", "~1.0")
            library = api.library
            data_item1_ref = library.data_items[0]
            data_item2_ref = library.data_items[1]
            # first data item gets displayed because there is an empty display panel.
            self.assertEqual(api.application.document_windows[0].display_data_item(data_item1_ref)._display_panel, display_panel)
            # the display is already filled. display panel should be None.
            self.assertIsNone(api.application.document_windows[0].display_data_item(data_item2_ref))
            # redisplay returns existing display panel.
            self.assertEqual(api.application.document_windows[0].display_data_item(data_item1_ref)._display_panel, display_panel)

    def test_lookup_unknown_instrument_or_hardware_source_returns_none(self):
        api = Facade.get_api("~1.0", "~1.0")
        self.assertIsNone(api.get_hardware_source_by_id("nonexistent_hardware", "~1.0"))
        self.assertIsNone(api.get_instrument_by_id("nonexistent_instrument", "~1.0"))

    def test_create_data_item_from_data_copies_data(self):
        memory_persistent_storage_system = DocumentModel.MemoryStorageSystem()
        document_model = DocumentModel.DocumentModel(persistent_storage_system=memory_persistent_storage_system)
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            api = Facade.get_api("~1.0", "~1.0")
            data = numpy.random.randn(2, 2)
            data_item = api.library.create_data_item_from_data(data)
            data[:, :] = numpy.random.randn(2, 2)
            self.assertFalse(numpy.array_equal(data, data_item.data))

    def test_create_data_item_from_xdata_copies_data(self):
        memory_persistent_storage_system = DocumentModel.MemoryStorageSystem()
        document_model = DocumentModel.DocumentModel(persistent_storage_system=memory_persistent_storage_system)
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            api = Facade.get_api("~1.0", "~1.0")
            data = numpy.random.randn(2, 2)
            xdata = api.create_data_and_metadata_from_data(data)
            data_item = api.library.create_data_item_from_data_and_metadata(xdata)
            data[:, :] = numpy.random.randn(2, 2)
            self.assertFalse(numpy.array_equal(data, data_item.data))

    def test_create_empty_data_item_and_set_data_copies_data(self):
        memory_persistent_storage_system = DocumentModel.MemoryStorageSystem()
        document_model = DocumentModel.DocumentModel(persistent_storage_system=memory_persistent_storage_system)
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            api = Facade.get_api("~1.0", "~1.0")
            data = numpy.random.randn(2, 2)
            data_item = api.library.create_data_item()
            data_item.set_data(data)
            data[:, :] = numpy.random.randn(2, 2)
            self.assertFalse(numpy.array_equal(data, data_item.data))

    def test_create_empty_data_item_and_set_xdata_copies_data(self):
        memory_persistent_storage_system = DocumentModel.MemoryStorageSystem()
        document_model = DocumentModel.DocumentModel(persistent_storage_system=memory_persistent_storage_system)
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            api = Facade.get_api("~1.0", "~1.0")
            data = numpy.random.randn(2, 2)
            data_item = api.library.create_data_item()
            data_item.xdata = api.create_data_and_metadata_from_data(data)
            data[:, :] = numpy.random.randn(2, 2)
            self.assertFalse(numpy.array_equal(data, data_item.data))

    class Computation1:
        def __init__(self, computation, **kwargs):
            self.computation = computation

        def execute(self, src):
            self.__src = src

        def commit(self):
            graphic = self.computation.get_result("graphic")
            if not graphic:
                graphic = self.__src.add_point_region(0.5, 0.5)
                self.computation.set_result("graphic", graphic)

    def test_register_library_computation_and_execute_it(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            api = Facade.get_api("~1.0")
            api.register_computation_type("computation1", self.Computation1)
            data = numpy.random.randn(2, 2)
            data_item = api.library.create_data_item()
            data_item.set_data(data)
            api.library.create_computation("computation1", inputs={"src": data_item}, outputs={"graphic": None})
            document_model.recompute_all()
            self.assertEqual(len(data_item.graphics), 1)

    class Computation2:
        def __init__(self, computation, **kwargs):
            self.computation = computation

        def execute(self, src, value):
            self.__src = src
            self.__value = value

        def commit(self):
            graphic = self.computation.get_result("graphic")
            if not graphic:
                graphic = self.__src.add_point_region(0.5, 0.5)
                self.computation.set_result("graphic", graphic)
            graphic.label = str(self.__value)

    def test_register_library_computation_with_variable_and_execute_it(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            api = Facade.get_api("~1.0")
            api.register_computation_type("computation2", self.Computation2)
            data = numpy.random.randn(2, 2)
            data_item = api.library.create_data_item()
            data_item.set_data(data)
            computation = api.library.create_computation("computation2", inputs={"src": data_item, "value": "label"}, outputs={"graphic": None})
            document_model.recompute_all()
            self.assertEqual(len(data_item.graphics), 1)
            self.assertEqual(data_item.graphics[0].label, "label")
            computation.set_input_value("value", "label2")
            document_model.recompute_all()
            self.assertEqual(len(data_item.graphics), 1)
            self.assertEqual(data_item.graphics[0].label, "label2")

    class Computation3:
        def __init__(self, computation, **kwargs):
            self.computation = computation

        def execute(self, src):
            self.__data = src.data
            self.__src = src

        def commit(self):
            dst = self.computation.get_result("dst")
            dst.set_data(self.__data)

    def test_library_computation_change_object_input(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            api = Facade.get_api("~1.0")
            api.register_computation_type("computation3", self.Computation3)
            data1 = numpy.ones((2, 2)) * 1
            data_item1 = api.library.create_data_item()
            data_item1.set_data(data1)
            data2 = numpy.ones((2, 2)) * 2
            data_item2 = api.library.create_data_item()
            data_item2.set_data(data2)
            dst_data_item = api.library.create_data_item()
            computation = api.library.create_computation("computation3", inputs={"src": data_item1}, outputs={"dst": dst_data_item})
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(dst_data_item.data, data1))
            computation.set_input_value("src", data_item2)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(dst_data_item.data, data2))

    class Computation4:
        def __init__(self, computation, **kwargs):
            self.computation = computation

        def execute(self, src):
            assert isinstance(src, Facade.DataSource)
            self.__data = src.data
            self.__src = src

        def commit(self):
            dst = self.computation.get_result("dst")
            dst.set_data(self.__data)

    def test_library_computation_change_object_data_source_input(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = self.app.create_document_controller(document_model, "library")
        with contextlib.closing(document_controller):
            api = Facade.get_api("~1.0")
            api.register_computation_type("computation4", self.Computation4)
            data1 = numpy.ones((2, 2)) * 1
            data_item1 = api.library.create_data_item()
            data_item1.set_data(data1)
            data2 = numpy.ones((2, 2)) * 2
            data_item2 = api.library.create_data_item()
            data_item2.set_data(data2)
            dst_data_item = api.library.create_data_item()
            computation = api.library.create_computation("computation4", inputs={"src": {"object": data_item1, "type": "data_source"}}, outputs={"dst": dst_data_item})
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(dst_data_item.data, data1))
            computation.set_input_value("src", {"object": data_item2, "type": "data_source"})
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(dst_data_item.data, data2))


if __name__ == '__main__':
    unittest.main()



import colorsys
import numpy
import math
import typing

def interpolate_colors(array: typing.List[int], x: int) -> typing.List[int]:
    """
    Creates a color map for values in array
    :param array: color map to interpolate
    :param x: number of colors
    :return: interpolated color map
    """
    out_array = []
    for i in range(x):
        if i % (x / (len(array) - 1)) == 0:
            index = i / (x / (len(array) - 1))
            out_array.append(array[int(index)])
        else:
            start_marker = array[math.floor(i / (x / (len(array) - 1)))]
            stop_marker = array[math.ceil(i / (x / (len(array) - 1)))]
            interp_amount = i % (x / (len(array) - 1)) / (x / (len(array) - 1))
            interp_color = numpy.rint(start_marker + ((stop_marker - start_marker) * interp_amount))
            out_array.append(interp_color)
    out_array[-1] = array[-1]
    return numpy.array(out_array).astype(numpy.uint8)

def generate_lookup_array_grayscale():
    out_list = []
    for i in range(256):
        out_list.append([i, i, i])
    return numpy.array(out_list)

lookup_arrays = {
    'magma':     [[0, 0, 0],
                  [127, 0, 127],
                  [96, 96, 255],
                  [64, 192, 255],
                  [223, 255, 255]],
    'viridis':   [[32, 32, 64],
                  [160, 160, 64],
                  [208, 208, 0],
                  [144, 255, 32],
                  [0, 255, 255]],
    'plasma':    [[127, 0, 0],
                  [127, 0, 127],
                  [127, 127, 255],
                  [127, 255, 255],
                  [195, 255, 255]],
    'ice':       [[0, 0, 0],
                  [127, 0, 0],
                  [255, 127, 0],
                  [255, 255, 127],
                  [255, 255, 255]]
}

def generate_lookup_array(color_map_id: str) -> numpy.array:
    return interpolate_colors(numpy.array(lookup_arrays[color_map_id]), 256)

def generate_lookup_array_hsv() -> numpy.array:
    result_array = []
    for lookup_value in range(256):
        color_values = [x * 255 for x in colorsys.hsv_to_rgb(lookup_value / 300, 1.0, 1.0)]
        result_array.append(color_values)
    return numpy.array(result_array).astype(int)

color_maps = {}
color_maps['magma'] = generate_lookup_array('magma')
color_maps['ice'] = generate_lookup_array('ice')
color_maps['plasma'] = generate_lookup_array('plasma')
color_maps['viridis'] = generate_lookup_array('viridis')
color_maps['hsv'] = generate_lookup_array_hsv()
color_maps['grayscale'] = generate_lookup_array_grayscale()


# futures
from __future__ import absolute_import

# standard libraries
import unittest

# third party libraries
# None

# local libraries
# None


class TestTaskClass(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_none(self):
        pass
